/*
    Fintek F81504 / F81508 / F81512

    Tested on
        Linux Kernel 3.4.5
*/

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/tty.h>
#include <linux/serial_reg.h>
#include <linux/serial_core.h>
#include <linux/8250_pci.h>
#include <linux/bitops.h>
#include <linux/serial_8250.h>
#include <linux/version.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <linux/mutex.h>

#include <linux/serial_8250.h>
#include <linux/serial_reg.h>
#include <linux/dmaengine.h>
#include <linux/gpio.h>

#define FINTEK_DEBUG 0

#define LOG_MESSAGE(x, y,...)	\
    do{ if(FINTEK_DEBUG) printk(x  y,##__VA_ARGS__ ); }while(0)

#define MAX_FINTEK_SERIAL_PORT 12
#define MAX_FINTEK_SERIAL_GPIO 6
#define MULTI_FUNCTION_PORT_FLAG -1

#define GPIO_MODE_OPENDRAIN
#define AER_FORCE_CHECK

#if 0
#define FORCE_INTO_GPIO_MODE
#endif

//#undef CONFIG_GPIOLIB

#define DRIVER_NAME "Fintek F81504/F81508/F81512"
#define GPIO_DISPLAY_NAME "F81504-Series-GPIO"

static u8 skip_port[MAX_FINTEK_SERIAL_GPIO] = { 2, 3, 8, 9, 10, 11 };

struct f81504_priv {
#ifdef CONFIG_GPIOLIB
	struct gpio_chip f81504_gpio_chip[6];
#endif
	struct mutex locker;
	int line_data[MAX_FINTEK_SERIAL_PORT];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	struct serial_rs485 rs485[MAX_FINTEK_SERIAL_PORT];
#endif
	u32 gpio_ioaddress;
	u8 multi_function;
};

#ifdef CONFIG_GPIOLIB
static int f81504_check_valid_gpio(struct gpio_chip *chip)
{
	int idx;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	for (idx = 0; idx < MAX_FINTEK_SERIAL_GPIO; ++idx) {
		if (priv->f81504_gpio_chip[idx].label
		    &&
		    (!strcmp(priv->f81504_gpio_chip[idx].label, chip->label)))
			break;
	}

	if (idx >= MAX_FINTEK_SERIAL_GPIO)
		return -EINVAL;

	return idx;
}

static int f81504_gpio_get(struct gpio_chip *chip, unsigned gpio_num)
{
	int idx;
	int tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);

	tmp = inb(priv->gpio_ioaddress + idx);

	return !!(tmp & (1 << gpio_num));

}

static int f81504_gpio_direction_in(struct gpio_chip *chip, unsigned gpio_num)
{
	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	mutex_lock(&priv->locker);

	/* set input mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08, tmp & ~(1 << gpio_num));

	mutex_unlock(&priv->locker);

	LOG_MESSAGE(KERN_INFO, "%s: gpio_num:%d, tmp:%d\n", __func__, gpio_num,
		    tmp);
	return 0;
}

static int f81504_gpio_direction_out(struct gpio_chip *chip,
				     unsigned gpio_num, int val)
{
	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);

	mutex_lock(&priv->locker);

	/* set output mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08, tmp | (1 << gpio_num));
#ifndef GPIO_MODE_OPENDRAIN
	/* set push pull mode */
	pci_read_config_byte(dev, 0xf8 + idx * 0x08 + 1, &tmp);
	pci_write_config_byte(dev, 0xf8 + idx * 0x08 + 1,
			      tmp | (1 << gpio_num));
#endif
	/* set output data */
	tmp = inb(priv->gpio_ioaddress + idx);

	if (val)
		outb(tmp | (1 << gpio_num), priv->gpio_ioaddress + idx);
	else
		outb(tmp & ~(1 << gpio_num), priv->gpio_ioaddress + idx);

	mutex_unlock(&priv->locker);

	return 0;
}

static void f81504_gpio_set(struct gpio_chip *chip, unsigned gpio_num, int val)
{
	LOG_MESSAGE(KERN_INFO, "%s: %d\n", __func__, __LINE__);
	f81504_gpio_direction_out(chip, gpio_num, val);
}

static int f81504_gpio_get_direction(struct gpio_chip *chip, unsigned offset)
{

	int idx;
	u8 tmp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	struct pci_dev *dev = container_of(chip->dev, struct pci_dev, dev);
#else
	struct pci_dev *dev = container_of(chip->parent, struct pci_dev, dev);
#endif
	struct f81504_priv *priv = pci_get_drvdata(dev);

	idx = f81504_check_valid_gpio(chip);
	if (idx < 0)
		return -EINVAL;

	mutex_lock(&priv->locker);
	pci_read_config_byte(dev, 0xf8 + idx * 0x08, &tmp);
	mutex_unlock(&priv->locker);

	/* F81504 bit value 0: input, 1: output 
	 * but GPIOLIB getdirection return 0: output, 1: input
	 * so we need to do a revert !
	 */
	LOG_MESSAGE(KERN_INFO, "%s: idx:%d, tmp:%d, offset:%d\n", __func__, idx,
		    !!(tmp & (1 << offset)), offset);
	return !(tmp & (1 << offset));

}

static struct gpio_chip f81504_gpio_chip_templete = {
	.owner = THIS_MODULE,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	.get_direction = f81504_gpio_get_direction,
#endif
	.get = f81504_gpio_get,
	.direction_input = f81504_gpio_direction_in,
	.set = f81504_gpio_set,
	.direction_output = f81504_gpio_direction_out,
	.base = -1,
};
#endif

int f81534_check_baudrate(unsigned int baud, int *idx)
{
	int index;
	u32 baudrate_table[3] = { 1500000, 1152000, 921600 };
	u32 quot[3];
	u32 rem[3];

	for (index = 0; index < sizeof(baudrate_table) / sizeof(int); ++index) {
		if (baud > baudrate_table[index])
			continue;

		quot[index] = DIV_ROUND_CLOSEST(baudrate_table[index], baud);
		rem[index] = baudrate_table[index] % baud;

		if (quot[index] && !rem[index]) {
			if (idx)
				*idx = index;
			return 0;
		}
	}

	return -1;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
void f81504_set_termios(struct uart_port *port, struct ktermios *termios,
			struct ktermios *old)
{
	struct pci_dev *dev = container_of(port->dev, struct pci_dev, dev);
	struct f81504_priv *priv = pci_get_drvdata(dev);
	unsigned int baud = tty_termios_baud_rate(termios);
	u32 baudrate_table[3] = { 1500000, 1152000, 921600 };
	u8 clock_table[3] = { 0x80, 0x40, 0xc0 };	/* clock + fifo enable */
	u8 tmp;
	int offset;
	int i;

	for (offset = 0; offset < MAX_FINTEK_SERIAL_PORT; ++offset)
		if ((priv->line_data[offset] != -1)
		    && (priv->line_data[offset] == port->line))
			break;

	if (offset == MAX_FINTEK_SERIAL_PORT) {
		dev_err(port->dev, "%s failed!!\n", __func__);
		return;
	}

	dev_dbg(port->dev, "%s: offset: %d\n", __func__, offset);

	do {
		pci_read_config_byte(dev, 0x40 + offset * 0x08, &tmp);

		if (baud <= 115200) {
			port->uartclk = 115200 * 16;
			pci_write_config_byte(dev, 0x40 + offset * 0x08,
					      tmp & ~(0xc0));
			break;
		}

		if (!f81534_check_baudrate(baud, &i)) {
			/* had optimize value */
			port->uartclk = baudrate_table[i] * 16;
			tmp = (tmp & ~(0xc0)) | clock_table[i];
			pci_write_config_byte(dev, 0x40 + offset * 0x08, tmp);
			dev_dbg(port->dev, "%s: optimized: %d\n", __func__, i);
			break;
		}

		if (old
		    && !f81534_check_baudrate(tty_termios_baud_rate(old),
					      NULL)) {
			baud = tty_termios_baud_rate(old);
			dev_dbg(port->dev, "%s: using old baud: %d\n", __func__,
				baud);
		} else {
			baud = 115200;
			dev_dbg(port->dev, "%s: force to baud: %d\n", __func__,
				baud);
		}

	} while (1);

	serial8250_do_set_termios(port, termios, old);

#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)

	struct uart_8250_port {
		struct uart_port	port;
		struct timer_list	timer;		/* "no irq" timer */
		struct list_head	list;		/* ports on this IRQ */
		unsigned short		capabilities;	/* port capabilities */
		unsigned short		bugs;		/* port bugs */
		unsigned int		tx_loadsz;	/* transmit fifo load size */
		unsigned char		acr;
		unsigned char		ier;
		unsigned char		lcr;
		unsigned char		mcr;
		unsigned char		mcr_mask;	/* mask of user bits */
		unsigned char		mcr_force;	/* mask of forced bits */
		unsigned char		cur_iotype;	/* Running I/O type */
	
		/*
		 * Some bits in registers are cleared on a read, so they must
		 * be saved whenever the register is read but the bits will not
		 * be immediately processed.
		 */
#define LSR_SAVE_FLAGS UART_LSR_BRK_ERROR_BITS
		unsigned char		lsr_saved_flags;
#define MSR_SAVE_FLAGS UART_MSR_ANY_DELTA
		unsigned char		msr_saved_flags;
	
		/* 8250 specific callbacks */
		int			(*dl_read)(struct uart_8250_port *);
		void			(*dl_write)(struct uart_8250_port *, int);
	};
#endif
	struct uart_8250_port *up =
		container_of(port, struct uart_8250_port, port);

	up->mcr_force |= UART_MCR_LOOP;
	//printk(KERN_INFO "%s: up->mcr: %x\n", __func__, up->mcr);
	//printk(KERN_INFO "%s: up->mcr_mask: %x\n", __func__, up->mcr_mask);
	//serial8250_do_set_mctrl(port, TIOCM_LOOP);
#endif
}

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
static int f81504_rs485_config(struct uart_port *port,
			       struct serial_rs485 *rs485)
{
	u8 setting = 0;
	int status;
	int index = (int)port->private_data;
	struct pci_dev *pci_dev = container_of(port->dev, struct pci_dev, dev);
	struct f81504_priv *priv = pci_get_drvdata(pci_dev);

	if (index < 0)
		return -EINVAL;

	status = pci_read_config_byte(pci_dev, 0x47 + 8 * index, &setting);
	if (status)
		return status;

	if (rs485->flags & SER_RS485_ENABLED) {
		setting |= BIT(4);

		if (rs485->flags & SER_RS485_RTS_ON_SEND)
			setting &= ~BIT(5);
		else
			setting |= BIT(5);
	} else {
		setting &= ~(BIT(4) | BIT(5));
	}

	status = pci_write_config_byte(pci_dev, 0x47 + 8 * index, setting);
	if (status)
		return status;

	port->rs485 = *rs485;
	priv->rs485[index] = *rs485;
	return 0;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
static int f81504_register_port(struct pci_dev *dev, int address, int idx)
{
	struct uart_port port;
	memset(&port, 0, sizeof(port));
	port.iotype = UPIO_PORT;
	port.mapbase = 0;
	port.membase = NULL;
	port.regshift = 0;
	port.irq = dev->irq;
	port.flags |= UPF_SKIP_TEST | UPF_FIXED_TYPE | UPF_SHARE_IRQ;
	port.uartclk = 1843200;
	port.dev = &dev->dev;
	port.iobase = address;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	port.set_termios = f81504_set_termios;
#endif
	port.type = PORT_16550A;
	port.fifosize = 32;
	port.private_data = (void *)idx;	/* save current idx */

	return serial8250_register_port(&port);
}
#else
static int f81504_register_port(struct pci_dev *dev, int address, int idx)
{

	struct uart_8250_port port;

	memset(&port, 0, sizeof(port));

	port.port.iotype = UPIO_PORT;
	port.port.mapbase = 0;
	port.port.membase = NULL;
	port.port.regshift = 0;
	port.port.irq = dev->irq;
	port.port.flags |= UPF_SKIP_TEST  | UPF_FIXED_TYPE | UPF_SHARE_IRQ;
	port.port.uartclk = 1843200;
	port.port.dev = &dev->dev;
	port.port.iobase = address;
	port.port.set_termios = f81504_set_termios;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	port.port.rs485_config = f81504_rs485_config;
#endif

	port.port.type = PORT_16550A;
	port.port.fifosize = 32;
	port.port.private_data = (void *)idx;	/* save current idx */

	return serial8250_register_8250_port(&port);
}

#endif

static int f81504_fill_port_register(struct pci_dev *dev, int idx,
				     unsigned long *ioport)
{
	struct pci_dev *pdev = dev;
	u32 bar_data[3];
	u8 config_base, valid;
	u8 msr, lsr;
	unsigned long iobase;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	u8 setting = 0;
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	struct serial_rs485 *rs485 = NULL;
#endif

	pci_read_config_dword(pdev, 0x24, &bar_data[0]);
	pci_read_config_dword(pdev, 0x20, &bar_data[1]);
	pci_read_config_dword(pdev, 0x1c, &bar_data[2]);

	if ((idx < 12) && (idx >= 0))
		config_base = 0x40 + idx * 0x08;
	else
		return -EINVAL;

	/* Calculate Real IO Port */
	iobase = (bar_data[idx / 4] & 0xffffffe0) + (idx % 4) * 8;

	/* Check is the port disabled by customer EEPROM */
	pci_read_config_byte(pdev, config_base + 0x00, &valid);
	//pr_info("idx: %d valid: %x\n", idx, valid);
	if ((valid & (BIT(2) | BIT(1))) == BIT(1)) {
		//pr_debug("%s: port idx: %d had disabled by EEPROM\n", __func__,
		//	 idx);
		return -ENODEV;
	}

	/* Enable UART I/O port */
	pci_write_config_byte(pdev, config_base + 0x00, 0x01);

	/* Select 128-byte FIFO and 8x FIFO threshold */
	pci_write_config_byte(pdev, config_base + 0x01, 0x33);

	/* LSB UART */
	pci_write_config_byte(pdev, config_base + 0x04, (u8) (iobase & 0xff));

	/* MSB UART */
	pci_write_config_byte(pdev, config_base + 0x05,
			      (u8) ((iobase & 0xff00) >> 8));

	/* force RS232. 9Bits Disable */
	pci_write_config_byte(pdev, 0x47 + 8 * idx, 1);

	/* irq number, this usually fails, but the spec says to do it anyway. */
	pci_write_config_byte(pdev, config_base + 0x06, pdev->irq);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 19, 0)
	rs485 = &priv->rs485[idx];

	if (rs485->flags & SER_RS485_ENABLED) {
		pci_read_config_byte(pdev, 0x40 + 8 * idx + 7, &setting);

		setting |= BIT(4);

		if (rs485->flags & SER_RS485_RTS_ON_SEND)
			setting &= ~BIT(5);
		else
			setting |= BIT(5);

		pci_write_config_byte(pdev, 0x40 + 8 * idx + 7, setting);
	}
#endif

	if (ioport != NULL)
		*ioport = iobase;

	/* Check Port exist */
	outb(0xc7, iobase + UART_FCR);
	outb(0x10, iobase + UART_MCR);
	inb(0x80);
	outb(0x00, iobase + UART_MCR);
	inb(0x80);
	udelay(60);		// delay

	lsr = inb(iobase + UART_LSR);
	msr = inb(iobase + UART_MSR);

	LOG_MESSAGE(KERN_INFO, "%s: port: %d lsr: %02x, msr: %02x\n", __func__,
		    idx, lsr, msr);

	/* Check H/W Port is exist */
	if ((lsr & UART_LSR_BI) && ((msr & 0xf0) == 0xf0))
		return -ENODEV;

	return 0;
}

static int f81504_is_gpio_set(int multi_function, int idx, int *set)
{
	int count;
	for (count = 0; count < MAX_FINTEK_SERIAL_GPIO; ++count) {
		if ((multi_function & (1 << count))
		    && (skip_port[count] == idx)) {

			if (set)
				*set = count;

			return 1;
		}
	}
	return 0;
}

static void f81504_init_gpio(struct pci_dev *dev)
{
	struct f81504_priv *priv = pci_get_drvdata(dev);
	u16 gpio_ioaddress;
	u8 tmp;

	pci_read_config_dword(dev, 0x18, &priv->gpio_ioaddress);
	priv->gpio_ioaddress &= 0xffffffe0;

	pci_read_config_word(dev, 0xf1, &gpio_ioaddress);

	if (gpio_ioaddress != priv->gpio_ioaddress) {
		pci_write_config_byte(dev, 0xf1,
				      (priv->gpio_ioaddress >> 0) & 0xff);
		pci_write_config_byte(dev, 0xf2,
				      (priv->gpio_ioaddress >> 8) & 0xff);
	}

	pci_read_config_byte(dev, 0xf0, &tmp);

	if (tmp != priv->multi_function) {
		/* Enable GPIO Port */
		pci_write_config_byte(dev, 0xf0, priv->multi_function);
	}
}

static int f81504_prepare_gpio(struct pci_dev *dev, int set)
{
#ifdef CONFIG_GPIOLIB
	struct f81504_priv *priv = pci_get_drvdata(dev);

	char *name = kmalloc(32, GFP_KERNEL);
	int rc;

	memcpy(&priv->f81504_gpio_chip[set],
	       &f81504_gpio_chip_templete, sizeof(f81504_gpio_chip_templete));

	sprintf(name, "%s%d", GPIO_DISPLAY_NAME, set);

	priv->f81504_gpio_chip[set].label = name;
	priv->f81504_gpio_chip[set].ngpio = 8;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	priv->f81504_gpio_chip[set].dev = &dev->dev;
#else
	priv->f81504_gpio_chip[set].parent = &dev->dev;
#endif
	rc = gpiochip_add(&priv->f81504_gpio_chip[set]);

	return rc;
#else
	return 0;
#endif
}

static void time_start(struct timespec *start)
{
	getnstimeofday(start);
}

static unsigned long long time_delta(struct timespec *start)
{
#define SEC 1000000000LL
	long long secs, usecs;
	struct timespec stop;

	getnstimeofday(&stop);
	
	secs = stop.tv_sec - start->tv_sec;
	usecs = stop.tv_nsec - start->tv_nsec;
	if (usecs < 0) {
		secs--;
		usecs += SEC;
	}

	return secs * SEC + usecs;
}

static int f81504_probe(struct pci_dev *dev, const struct pci_device_id *dev_id)
{
	int idx;
	int rc;
	int set;
	struct f81504_priv *priv = NULL;
	unsigned long iobase = 0;

	rc = pci_enable_device(dev);
	if (rc)
		return rc;
	
	priv = kzalloc(sizeof(struct f81504_priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	pci_set_drvdata(dev, priv);

	mutex_init(&priv->locker);
#ifdef FORCE_INTO_GPIO_MODE
	pci_write_config_byte(dev, 0xf3, 0x00);
#endif

	pci_read_config_byte(dev, 0xf3, &priv->multi_function);

	/* max multi function with 0x3f - 6devices */
	priv->multi_function = (~priv->multi_function) & 0x3f;

	dev_dbg(&dev->dev, "%s: Multifunction inv-Flag:%d\n",
		__func__, (int)priv->multi_function);

	for (idx = 0; idx < MAX_FINTEK_SERIAL_PORT; ++idx)
		priv->line_data[idx] = MULTI_FUNCTION_PORT_FLAG;

	f81504_init_gpio(dev);

	for (idx = 0; idx < dev_id->driver_data; ++idx) {
		if (f81504_is_gpio_set(priv->multi_function, idx, &set)) {

			rc = f81504_prepare_gpio(dev, set);
			if (rc) {
				LOG_MESSAGE(KERN_INFO,
					    "%s: gpiochip_add failed:%d\n",
					    __func__, rc);
			}
		} else {
			rc = f81504_fill_port_register(dev, idx, &iobase);
			if (rc < 0) {
				if (rc == -ENODEV)	/* port disbled */
					continue;

				printk(KERN_WARNING "port idx: %d,failed!!\n",
				       idx);
				continue;
			}

			priv->line_data[idx] =
			    f81504_register_port(dev, iobase, idx);

			dev_dbg(&dev->dev, "idx: %d\n", idx);
			dev_dbg(&dev->dev, "register with: %d\n",
				priv->line_data[idx]);
		}
	}

	if (dev_id->driver_data != 8)
		return 0;

	/* if F81508, 2multi + 4gpio sets need check */
	for (idx = dev_id->driver_data; idx < MAX_FINTEK_SERIAL_PORT; ++idx) {
		if (!f81504_is_gpio_set(priv->multi_function, idx, &set))
			continue;

		rc = f81504_prepare_gpio(dev, set);
		if (rc) {
			LOG_MESSAGE(KERN_INFO,
				    "%s: gpiochip_add failed:%d\n",
				    __func__, rc);
		}
	}

	dev->dev_flags |= PCI_DEV_FLAGS_NO_D3;
	return 0;
}

static void f81504_remove(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 18, 0)
	int rc;
#endif
	int i;

	LOG_MESSAGE(KERN_INFO, "f81504_remove\n");

	if (priv) {
		for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
			if (priv->line_data[i] > 0)
				serial8250_unregister_port(priv->line_data[i]);

#ifdef CONFIG_GPIOLIB
		for (i = 0; i < MAX_FINTEK_SERIAL_GPIO; ++i) {
			if (priv->f81504_gpio_chip[i].label) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 18, 0)
				gpiochip_remove(&priv->f81504_gpio_chip[i]);
#else
				rc = gpiochip_remove(&priv->f81504_gpio_chip
						     [i]);
				if (rc)
					LOG_MESSAGE(KERN_INFO,
						    "%s: gpiochip_remove:%d\n",
						    __func__, rc);
#endif
				kfree(priv->f81504_gpio_chip[i].label);
			}
		}
#endif

		kfree(priv);
	}

	pci_set_drvdata(dev, NULL);
	pci_disable_device(dev);

}

static int f81504_suspend(struct pci_dev *dev, pm_message_t state)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i;

	pci_save_state(dev);
	pci_set_power_state(dev, pci_choose_state(dev, state));

	if (!priv)
		return -ENODEV;

	for (i = 0; i < MAX_FINTEK_SERIAL_PORT; ++i)
		if (priv->line_data[i] > 0)
			serial8250_suspend_port(priv->line_data[i]);

	return 0;
}

static int f81504_resume(struct pci_dev *dev)
{
	struct f81504_priv *priv = (struct f81504_priv *)pci_get_drvdata(dev);
	int i, devices;
	int rc;

	pci_set_power_state(dev, PCI_D0);
	pci_restore_state(dev);

	if (!priv)
		return -ENODEV;

	switch (dev->device) {
	case 0x1104:		/* 4 ports */
	case 0x1108:		/* 8 ports */
		devices = dev->device & 0xff;
		break;
	case 0x1112:		/* 12 ports */
		devices = 12;
		break;
	default:
		return -EINVAL;
	}

	f81504_init_gpio(dev);

	for (i = 0; i < devices; ++i) {
		if (priv->line_data[i] > 0) {
			rc = f81504_fill_port_register(dev, i, NULL);
			if (rc < 0) {
				printk(KERN_INFO "%s: idx: %d failed\n",
				       __func__, i);
				//continue;
			}

			serial8250_resume_port(priv->line_data[i]);
		}
	}
	return 0;
}

static const struct pci_device_id f81504_dev_table[] = {
	/* Fintek PCI serial cards */
	{PCI_DEVICE(0x1c29, 0x1104),.driver_data = 4},
	{PCI_DEVICE(0x1c29, 0x1108),.driver_data = 8},
	{PCI_DEVICE(0x1c29, 0x1112),.driver_data = 12},
	{}
};

MODULE_DEVICE_TABLE(pci, f81504_dev_table);

static struct pci_driver f81504_driver = {
	.name = "f81504_pci2uart-4-8-12",
	.probe = f81504_probe,
	.remove = f81504_remove,
#ifdef CONFIG_PM
	.suspend = f81504_suspend,
	.resume = f81504_resume,
#endif
	.id_table = f81504_dev_table,
};

static __init int f81504_init(void)
{
	return pci_register_driver(&f81504_driver);
}

static __exit void f81504_exit(void)
{
	pci_unregister_driver(&f81504_driver);
}

module_init(f81504_init);
module_exit(f81504_exit);

MODULE_DESCRIPTION(DRIVER_NAME);
MODULE_AUTHOR("Peter Hong <Peter_Hong@fintek.com.tw>");
MODULE_LICENSE("GPL");
